;
; Error Program
;
; By Bryan Turner
;
; This program executes one error each execution to test the ability of the system
;    to recover from the errors.
;
	include "flib.h"
	include "tios.h"
	xdef	_main
	xdef	_comment

_comment:
	dc.b	"Error Testor",0
	ds.w	0

_main:
	jsr	flib::clr_scr
	
	move.w	#1,-(a7)
	jsr	tios::FontSetSys	; 8pt font
	addq	#2,a7

	move.w	#4,-(a7)			; Print 'Any Key' string..
	pea	test_str(pc)
	clr.l	-(a7)
	jsr	tios::DrawStrXY
	lea	10(a7),a7

	lea	errstr,a0
	move.w	errorNum(pc),d0			; Load error number
	move.w	d0,d7				; Save a copy
	lsl.w	#1,d0
	move.w	0(a0,d0.w),d0			; Load error string offset

	addq	#1,errorNum			; Increment error number
	and.w	#$F,errorNum			; Mod 16, wraps around after last test

	move.w	#4,-(a7)			; Print the error we're about to execute
	pea	0(a0,d0.w)
	move.l	#$00100010,-(a7)
	jsr	tios::DrawStrXY
	lea	10(a7),a7

	jsr	flib::idle_loop			; Wait for a key before proceeding...

	lsl.w	#1,d7
	lea	jump_tab(pc),a0			; Load jumptable address
	move.w	0(a0,d7.w),d0			; Load offset from jumptable
	jsr	0(a0,d0.w)			; Jump to code!
	rts

jump_tab:
	dc.w	dobus-jump_tab
	dc.w	doaddr-jump_tab
	dc.w	doill-jump_tab
	dc.w	dodivide-jump_tab
	dc.w	dochkx-jump_tab
	dc.w	dotrapvx-jump_tab
	dc.w	dopriv-jump_tab
	dc.w	dotrace-jump_tab
	dc.w	dofline-jump_tab
	dc.w	dospurr-jump_tab
	dc.w	doauto3-jump_tab
	dc.w	doauto7-jump_tab
	dc.w	dotrapb-jump_tab
	dc.w	dotrapc-jump_tab
	dc.w	dotrapd-jump_tab
	dc.w	dotrape-jump_tab

;
; --------------------------------------------
; The meat of the code: Creating errors!  Yay!
;

; This procedure is used to acquire supervisor mode from user mode
; On entry: A0 = procedure to run once supervisor mode is entered
; Must exit from supervisor code with RTE!!
getsuper:
	bclr.b	#2,($600001)		; Memory protect OFF

	move.l	($b8),d0		; Load old TRAP #$E processor
	lea	getsup(pc),a1		; Load our proc
	move.l	a1,($b8)		; Store new TRAP #$E processor
	trap	#$e
	rts
getsup:
	move.l	d0,($b8)		; Reset TRAP #$E processor
	bset.b	#2,($600001)		; Memory protect ON
	jmp	(a0)			; Jump to proc

dobus:
	; Bus error caused by... Sorry, can't think of a clean way to induce one.. :(
	rts

doaddr:
	; Address Error: word write on odd-byte boundary
	lea	junk(pc),a0
	addq	#1,a0
	move.w	#$1234,(a0)		; Oops! ;)
	rts

doill:
	; Illegal instruction
	illegal				; Pretty simple eh?
	rts

dodivide:
	; Divide by zero trap
	moveq	#1,d0
	moveq	#0,d1
	divu	d1,d0			; Oh, did I do that?
	rts

dochkx:
	; Check Exception
	moveq	#4,d0
	chk.l	#2,d0
	rts

dotrapvx:
	; TRAPV Vector
	lea	trapvexec(pc),a0	; Proc to run in supervisor mode..
	jsr 	getsuper		; Call supervisor mode function
	trapv
	rts
trapvexec:
	or.w	#$0002,(a7)		; Turn on overflow bit manually
	rte				; Nothing else seemed to work.. I must be stupid..

dopriv:
	; Privelege violation
	move.w	#2700,sr		; Attempt supervisor instruction from user mode
	rts

dotrace:
	; Trace trap: a bit more tricky since we need supervisor mode
	lea	traceon(pc),a0
	jsr	getsuper
	rts

traceon:
	or.w	#$8000,(a7)		; Turn trace bit on for return SR
	rte				; Return and cause trace exception

dofline:
	; F-Line Exception
	dc.w	$F001
	rts
	
dospurr:
	; Spurrious Interrupt: Sorry can't think of a way to cause it :(
	rts

doauto3:
	; Method to cause Auto-Int 3 discovered by Tleilax:
	;move.b	#0,($600015)
	;
	; BT: Not validated, I couldn't get it to work.. :(
	;     Anyone else know how?
	;
	rts

doauto7:
	; Memory Protect Violation: write to low memory!
	bset.b	#2,($600001)		; Memory protect ON
	lea	$30,a0
	move.l	#-1,(a0)		; Write to low memory!
	rts

	;
	; The trap functions are pretty simple, but included for completeness...
	;
dotrapb:
	trap	#$b
	rts
dotrapc:
	trap	#$c
	rts
dotrapd:
	trap	#$d
	rts
dotrape:
	trap	#$e
	rts


; The error to be tested this go-around
errorNum:
	dc.w	0

junk:
	dc.l	0,0

test_str:
	dc.b	'Any key to test error:',0
	EVEN

;
; Errors tested, and their order:
;
errstr:
	dc.w	bus-errstr
	dc.w	addr-errstr
	dc.w	ill-errstr
	dc.w	divide-errstr
	dc.w	chkx-errstr
	dc.w	trapvx-errstr
	dc.w	priv-errstr
	dc.w	trace-errstr
	dc.w	fline-errstr
	dc.w	spurr-errstr
	dc.w	auto3-errstr
	dc.w	auto7-errstr
	dc.w	trapb-errstr
	dc.w	trapc-errstr
	dc.w	trapd-errstr
	dc.w	trape-errstr
bus:
	dc.b 'Bus Error - not complete',0
	EVEN
addr:
	dc.b 'Address Error',0
	EVEN
ill:
	dc.b 'Illegal Instruction',0
	EVEN
divide:
	dc.b 'Divide by Zero',0
	EVEN
chkx:
	dc.b 'CHK Exception',0
	EVEN
trapvx:
	dc.b 'TRAPV Exception',0
	EVEN
priv:
	dc.b 'Privelege Violation',0
	EVEN
trace:
	dc.b 'Trace Trap',0
	EVEN
fline:
	dc.b 'F-Line Trap',0
	EVEN
spurr:
	dc.b 'Spurrious Interrupt - not complete',0
	EVEN
auto3:
	dc.b 'Auto-Int 3 - not complete',0
	EVEN
auto7:
	dc.b 'Auto-Int 7',0
	EVEN
trapb:
	dc.b 'TRAP $B',0
	EVEN
trapc:
	dc.b 'TRAP $C',0
	EVEN
trapd:
	dc.b 'TRAP $D',0
	EVEN
trape:
	dc.b 'TRAP $E',0
	EVEN

	end